#ifdef 	MODVERSIONS
#ifndef MODULE
#define MODULE
#endif
#endif

#ifdef	MODULE
#include <linux/config.h>
#ifdef	MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/module.h>
#else
#define	MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif


//#include <linux/config.h>
//#include <linux/module.h>	/* For EXPORT_SYMBOL */
#include <linux/version.h>
#include <linux/autoconf.h>

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/fcntl.h>
#include <linux/string.h>
#include <linux/major.h>
#include <linux/mm.h>
#include <linux/init.h>
#include <linux/devfs_fs_kernel.h>

#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/bitops.h>

#define BUILDING_PTY_C 1
#include <linux/devpts_fs.h>

#include "vpty.h"

#ifndef CONFIG_LEGACY_PTY_COUNT
#define CONFIG_LEGACY_PTY_COUNT	256
#endif

#ifndef VPTY_COUNT	
#define VPTY_COUNT NR_PTYS
#endif

#define VPTY_NR_MAJORS	(VPTY_COUNT/NR_PTYS)

#ifndef VPTY_MASTER_MAJOR
#define VPTY_MASTER_MAJOR	100
#endif

#ifndef VPTY_SLAVE_MAJOR
#define VPTY_SLAVE_MAJOR	101
#endif

#define	VERSION_CODE(ver,rel,seq)	((ver << 16) | (rel << 8) | seq)

#if (LINUX_VERSION_CODE < VERSION_CODE(2,1,0))
#define copy_from_user memcpy_fromfs
#define copy_to_user memcpy_tofs
#define put_to_user(arg1, arg2) put_fs_long(arg1, (unsigned long *)arg2)
#define get_from_user(arg1, arg2) arg1 = get_fs_long((unsigned long *)arg2)
#define schedule_timeout(x) {current->timeout = jiffies + (x); schedule();}
#define signal_pending(x) ((x)->signal & ~(x)->blocked)
#else
#include <asm/uaccess.h>
#define put_to_user(arg1, arg2) put_user(arg1, (unsigned long *)arg2)
#define get_from_user(arg1, arg2) get_user(arg1, (unsigned int *)arg2)
#endif 

#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))

struct vpty_struct {
	int	magic;
	wait_queue_head_t open_wait;
};

//#define PTY_MAGIC 0x5001

static struct tty_driver vpty_driver[VPTY_NR_MAJORS];
static struct tty_driver vpty_slave_driver[VPTY_NR_MAJORS];
static int vpty_refcount[VPTY_NR_MAJORS];

/* Note: one set of tables for BSD and one for Unix98 */
static struct tty_struct *vpty_table[VPTY_NR_MAJORS][NR_PTYS];
static struct termios *vpty_termios[VPTY_NR_MAJORS][NR_PTYS];
static struct termios *vpty_termios_locked[VPTY_NR_MAJORS][NR_PTYS];
static struct tty_struct *vttyp_table[VPTY_NR_MAJORS][NR_PTYS];
static struct termios *vttyp_termios[VPTY_NR_MAJORS][NR_PTYS];
static struct termios *vttyp_termios_locked[VPTY_NR_MAJORS][NR_PTYS];
static struct vpty_struct vpty_state[VPTY_NR_MAJORS][NR_PTYS];

#else
static struct tty_driver *vpty_driver, *vpty_slave_driver;
#endif

#ifdef MODULE
MODULE_AUTHOR("HTJ");
MODULE_DESCRIPTION("Vpty driver");
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif
#endif

#define MIN(a,b)	((a) < (b) ? (a) : (b))
static int verbose = 1;

static void vpty_close(struct tty_struct * tty, struct file * filp)
{
	if (!tty)
		return;
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
	if (tty->driver.subtype == PTY_TYPE_MASTER) {
#else
	if (tty->driver->subtype == PTY_TYPE_MASTER) {
#endif
		if (tty->count > 1)
			printk("master vpty_close: count = %d!!\n", tty->count);
	} else {
		if (tty->count > 2)
			return;
	}
	wake_up_interruptible(&tty->read_wait);
	wake_up_interruptible(&tty->write_wait);
	tty->packet = 0;
	if (!tty->link)
		return;
	tty->link->packet = 0;
	wake_up_interruptible(&tty->link->read_wait);
	wake_up_interruptible(&tty->link->write_wait);
	set_bit(TTY_OTHER_CLOSED, &tty->link->flags);
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
	if (tty->driver.subtype == PTY_TYPE_MASTER) {
#else
	if (tty->driver->subtype == PTY_TYPE_MASTER) {
#endif
		set_bit(TTY_OTHER_CLOSED, &tty->flags);
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
		tty_unregister_devfs (&tty->link->driver, MINOR (tty->device));
#endif
//		tty_vhangup(tty->link);
		tty_hangup(tty->link);
	}
}

static void vpty_unthrottle(struct tty_struct * tty)
{
	struct tty_struct *o_tty = tty->link;

	if (!o_tty)
		return;

	if ((o_tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
	    o_tty->ldisc.write_wakeup)
		(o_tty->ldisc.write_wakeup)(o_tty);
	wake_up_interruptible(&o_tty->write_wait);
	set_bit(TTY_THROTTLED, &tty->flags);
}

static int vpty_write(struct tty_struct * tty, int from_user, const unsigned char *buf, int count)
{
	struct tty_struct *to = tty->link;
	int	c=0, n, room;
	char	*temp_buffer;

	if (!to || tty->stopped)
		return 0;

	if (from_user) {
		down(&tty->flip.pty_sem);
		temp_buffer = &tty->flip.char_buf[0];
		while (count > 0) {
			/* check space so we don't copy needlessly */ 
			n = to->ldisc.receive_room(to);
			if (n > count)
				n = count;
			if (!n) break;

			n  = MIN(n, PTY_BUF_SIZE);
			n -= copy_from_user(temp_buffer, buf, n);
			if (!n) {
				if (!c)
					c = -EFAULT;
				break;
			}

			/* check again in case the buffer filled up */
			room = to->ldisc.receive_room(to);
			if (n > room)
				n = room;
			if (!n) break;
			buf   += n; 
			c     += n;
			count -= n;
			to->ldisc.receive_buf(to, temp_buffer, 0, n);
		}
		up(&tty->flip.pty_sem);
	} else {
		c = to->ldisc.receive_room(to);
		if (c > count)
			c = count;
		to->ldisc.receive_buf(to, buf, 0, c);
	}
	
	return c;
}

static int vpty_write_room(struct tty_struct *tty)
{
	struct tty_struct *to = tty->link;

	if (!to || tty->stopped)
		return 0;

	return to->ldisc.receive_room(to);
}

static int vpty_chars_in_buffer(struct tty_struct *tty)
{
	struct tty_struct *to = tty->link;
	int count;

	if (!to || !to->ldisc.chars_in_buffer)
		return 0;

	/* The ldisc must report 0 if no characters available to be read */
	count = to->ldisc.chars_in_buffer(to);

#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
	if (tty->driver.subtype == PTY_TYPE_SLAVE) return count;
#else
	if (tty->driver->subtype == PTY_TYPE_SLAVE) return count;
#endif

	/* Master side driver ... if the other side's read buffer is less than 
	 * half full, return 0 to allow writers to proceed; otherwise return
	 * the count.  This leaves a comfortable margin to avoid overflow, 
	 * and still allows half a buffer's worth of typed-ahead commands.
	 */
	return ((count < N_TTY_BUF_SIZE/2) ? 0 : count);
}

static int vpty_set_lock(struct tty_struct *tty, int * arg)
{
	int val;
	if (get_user(val,arg))
		return -EFAULT;
	if (val)
		set_bit(TTY_PTY_LOCK, &tty->flags);
	else
		clear_bit(TTY_PTY_LOCK, &tty->flags);
	return 0;
}

static int vpty_bsd_ioctl(struct tty_struct *tty, struct file *file,
			unsigned int cmd, unsigned long arg)
{
	if (!tty) {
		printk("vpty_ioctl called with NULL tty!\n");
		return -EIO;
	}
	switch(cmd) {
	case TIOCSPTLCK: /* Set PT Lock (disallow slave open) */
		return vpty_set_lock(tty, (int *) arg);
	}
	return -ENOIOCTLCMD;
}

static void vpty_flush_buffer(struct tty_struct *tty)
{
	struct tty_struct *to = tty->link;
	
	if (!to)
		return;
	
	if (to->ldisc.flush_buffer)
		to->ldisc.flush_buffer(to);
	
	if (to->packet) {
		tty->ctrl_status |= TIOCPKT_FLUSHWRITE;
		wake_up_interruptible(&to->read_wait);
	}
}

static int vpty_open(struct tty_struct *tty, struct file * filp)
{
	int	retval = -ENODEV;
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
	int	line;
	struct	vpty_struct *pty;
#endif

	if (!tty || !tty->link)
		goto out;


#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
	line = MINOR(tty->device) - tty->driver.minor_start;
	if ((line < 0) || (line >= NR_PTYS))
		goto out;
	pty = (struct vpty_struct *)(tty->driver.driver_state) + line;
	tty->driver_data = pty;
#endif 

	retval = -EIO;
	if (test_bit(TTY_OTHER_CLOSED, &tty->flags))
		goto out;
	if (test_bit(TTY_PTY_LOCK, &tty->link->flags))
		goto out;
	if (tty->link->count != 1)
		goto out;

	clear_bit(TTY_OTHER_CLOSED, &tty->link->flags);
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
	wake_up_interruptible(&pty->open_wait);
#endif
	set_bit(TTY_THROTTLED, &tty->flags);

	/*  Register a slave for the master  */
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
	if (tty->driver.major == VPTY_MASTER_MAJOR)
		tty_register_devfs(&tty->link->driver,
				   DEVFS_FL_CURRENT_OWNER | DEVFS_FL_WAIT,
				   tty->link->driver.minor_start +
				   MINOR(tty->device)-tty->driver.minor_start);
#else
	set_bit(TTY_DO_WRITE_WAKEUP, &tty->flags);
#endif

	retval = 0;
out:
	return retval;
}

static void vpty_set_termios(struct tty_struct *tty, struct termios *old_termios)
{
        tty->termios->c_cflag &= ~(CSIZE | PARENB);
        tty->termios->c_cflag |= (CS8 | CREAD);
}


#if (LINUX_VERSION_CODE >= VERSION_CODE(2,6,0))
static struct tty_operations vpty_ops = {
	.open = vpty_open,
	.close = vpty_close,
	.write = vpty_write,
	.write_room = vpty_write_room,
	.flush_buffer = vpty_flush_buffer,
	.chars_in_buffer = vpty_chars_in_buffer,
	.unthrottle = vpty_unthrottle,
	.set_termios = vpty_set_termios,
};
#endif

#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
#define DRV_VAR(n)		(&vpty_driver[n])
#define DRV_VAR_S(n)		(&vpty_slave_driver[n])
#define DRV_VAR_P(n,x)		vpty_driver[n].x
#define DRV_VAR_SP(n,x)		vpty_slave_driver[n].x
#else
#define DRV_VAR			(vpty_driver)
#define DRV_VAR_S		(vpty_slave_driver)
#define DRV_VAR_P(x)	vpty_driver->x
#define DRV_VAR_SP(x)	vpty_slave_driver->x
#endif

static int vpty_init(void)
{
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
	int i, n;
	int master, slave;

	/* Traditional BSD devices */
	for(n = 0; n < VPTY_NR_MAJORS; n++)
	{
		master = VPTY_MASTER_MAJOR + n * 2;
		slave = VPTY_SLAVE_MAJOR + n * 2;

		printk("Register vpty pair device,  master  %d, slave %d\n", master, slave);
		
		memset(&vpty_state[n], 0, sizeof(vpty_state[n]));
		for (i = 0; i < NR_PTYS; i++)
			init_waitqueue_head(&vpty_state[n][i].open_wait);
		
		memset(&vpty_driver[n], 0, sizeof(struct tty_driver));

		DRV_VAR_P(n,magic) = TTY_DRIVER_MAGIC;
	#ifdef CONFIG_DEVFS_FS
		DRV_VAR_P(n,name) = "vpty/m%d";
	#else
		DRV_VAR_P(n,name) = "vpty";
	#endif
		DRV_VAR_P(n,num) = NR_PTYS;
		DRV_VAR_P(n,refcount) = &vpty_refcount[n];
		DRV_VAR_P(n,table) = vpty_table[n];
		DRV_VAR_P(n, termios) = vpty_termios[n];
		DRV_VAR_P(n, termios_locked) = vpty_termios_locked[n];
		DRV_VAR_P(n, driver_state) = vpty_state[n];
		DRV_VAR_P(n, other) = &vpty_slave_driver[n];

		DRV_VAR_P(n,driver_name) = "vpty_master";
		DRV_VAR_P(n, major) = master;
		DRV_VAR_P(n, minor_start) = 0;
		DRV_VAR_P(n, type) = TTY_DRIVER_TYPE_PTY;
		DRV_VAR_P(n, subtype) = PTY_TYPE_MASTER;
		DRV_VAR_P(n, init_termios) = tty_std_termios;
		DRV_VAR_P(n, init_termios.c_iflag) = 0;
		DRV_VAR_P(n, init_termios.c_oflag) = 0;
		DRV_VAR_P(n, init_termios.c_cflag) = B38400 | CS8 | CREAD;
		DRV_VAR_P(n, init_termios.c_lflag) = 0;
		DRV_VAR_P(n, flags) = TTY_DRIVER_RESET_TERMIOS | TTY_DRIVER_REAL_RAW;
		DRV_VAR_P(n, open) = vpty_open;
		DRV_VAR_P(n, close) = vpty_close;
		DRV_VAR_P(n, write) = vpty_write;
		DRV_VAR_P(n, write_room) = vpty_write_room;
		DRV_VAR_P(n, flush_buffer) = vpty_flush_buffer;
		DRV_VAR_P(n, chars_in_buffer) = vpty_chars_in_buffer;
		DRV_VAR_P(n, unthrottle) = vpty_unthrottle;
		DRV_VAR_P(n, set_termios) = vpty_set_termios;

		vpty_slave_driver[n] = vpty_driver[n];

		DRV_VAR_SP(n, proc_entry) = 0;
	#ifdef CONFIG_DEVFS_FS
		DRV_VAR_SP(n, name) = "vpty/s%d";
	#else
		DRV_VAR_SP(n, name) = "vttyp";
	#endif
		DRV_VAR_SP(n, flags) |= TTY_DRIVER_NO_DEVFS;
		DRV_VAR_SP(n, table) = vttyp_table[n];
		DRV_VAR_SP(n, termios) = vttyp_termios[n];
		DRV_VAR_SP(n, termios_locked) = vttyp_termios_locked[n];
		DRV_VAR_SP(n, driver_state) = vpty_state[n];
		DRV_VAR_SP(n, other) = &vpty_driver[n];
		DRV_VAR_SP(n, driver_name) = "vpty_slave";
		DRV_VAR_SP(n, subtype) = PTY_TYPE_SLAVE;
		DRV_VAR_SP(n, major) = slave;
		DRV_VAR_SP(n, minor_start) = 0;
		DRV_VAR_SP(n, init_termios) = tty_std_termios;
		DRV_VAR_SP(n, init_termios.c_cflag) = B38400 | CS8 | CREAD;

		if (tty_register_driver(DRV_VAR(n)))
			panic("Couldn't register pty driver");
		if (tty_register_driver(DRV_VAR_S(n)))
			panic("Couldn't register pty slave driver");

		vpty_driver[n].ioctl = vpty_bsd_ioctl;
	}

#else
	
	vpty_driver = alloc_tty_driver(NR_PTYS * VPTY_NR_MAJORS);
	if (!vpty_driver)
		panic("Couldn't allocate pty driver");

	vpty_slave_driver = alloc_tty_driver(NR_PTYS * VPTY_NR_MAJORS);
	if (!vpty_slave_driver)
		panic("Couldn't allocate pty slave driver");

	DRV_VAR_P(owner) = THIS_MODULE;
	DRV_VAR_P(name) = "vpty";
	DRV_VAR_P(devfs_name) = "vpty/m";
	DRV_VAR_P(other) = vpty_slave_driver;

	tty_set_operations(vpty_driver, &vpty_ops);
	vpty_driver->ioctl = vpty_bsd_ioctl;

	DRV_VAR_P(driver_name) = "vpty_master";
	DRV_VAR_P(major) = VPTY_MASTER_MAJOR;
	DRV_VAR_P(minor_start) = 0;
	DRV_VAR_P(type) = TTY_DRIVER_TYPE_PTY;
	DRV_VAR_P(subtype) = PTY_TYPE_MASTER;
	DRV_VAR_P(init_termios) = tty_std_termios;
	DRV_VAR_P(init_termios.c_iflag) = 0;
	DRV_VAR_P(init_termios.c_oflag) = 0;
	DRV_VAR_P(init_termios.c_cflag) = B38400 | CS8 | CREAD;
	DRV_VAR_P(init_termios.c_lflag) = 0;
	DRV_VAR_P(flags) = TTY_DRIVER_RESET_TERMIOS | TTY_DRIVER_REAL_RAW;

	DRV_VAR_SP(owner) = THIS_MODULE;
	DRV_VAR_SP(name) = "vttyp";
	DRV_VAR_SP(devfs_name) = "vpty/s";
	DRV_VAR_SP(type) = TTY_DRIVER_TYPE_PTY;
	DRV_VAR_SP(flags) = TTY_DRIVER_RESET_TERMIOS |
			TTY_DRIVER_REAL_RAW | TTY_DRIVER_NO_DEVFS;

	DRV_VAR_SP(other) = vpty_driver;
	tty_set_operations(vpty_slave_driver, &vpty_ops);
	DRV_VAR_SP(driver_name) = "vpty_slave";
	DRV_VAR_SP(subtype) = PTY_TYPE_SLAVE;
	DRV_VAR_SP(major) = VPTY_SLAVE_MAJOR;
	DRV_VAR_SP(minor_start) = 0;
	DRV_VAR_SP(init_termios) = tty_std_termios;
	DRV_VAR_SP(init_termios.c_cflag) = B38400 | CS8 | CREAD;

	if (tty_register_driver(DRV_VAR))
		panic("Couldn't register pty driver");
	if (tty_register_driver(DRV_VAR_S))
		panic("Couldn't register pty slave driver");

#endif

	return 0;
}

#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
int init_module(void)
#else
static int __init vpty_module_init(void)
#endif
{
	int	ret;
	int	vpty_cnt = VPTY_COUNT;

	if(verbose)
		printk("Loading vpty Driver , total vpty count %d...\n", vpty_cnt);
	ret = vpty_init();
	if(verbose)
		printk("Done.\n");
	return (ret);
}

#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
void cleanup_module(void)
#else
static void __exit vpty_module_exit(void)
#endif
{
	int n;
	if(verbose)
		printk("Unloading vpty Driver ...\n");

#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
	for(n = 0; n < VPTY_NR_MAJORS; n++)
	{
		tty_unregister_driver(DRV_VAR(n));
		tty_unregister_driver(DRV_VAR_S(n));
	}
#else
	n = 0;
	tty_unregister_driver(DRV_VAR);
	tty_unregister_driver(DRV_VAR_S);
#endif

	if(verbose)
		printk("Done.\n");
}

# if (LINUX_VERSION_CODE >= VERSION_CODE(2,6,0))
module_init(vpty_module_init);
module_exit(vpty_module_exit); 
#else
#ifdef MODULE
int init_module(void);
void cleanup_module(void);
#endif
#endif


